/* Copyright (c) 1997-2001 Guenter Geiger, Miller Puckette, Larry Troxler,
* Winfried Ritsch, Karl MacMillan, and others.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/* this file implements the sys_ functions profiled in m_imp.h for
 audio and MIDI I/O on Macintosh OS X. 
 
 Audio simply calls routines in s_portaudio.c, which in turn call the
 portaudio package.  s_portaudio.c is also intended for use from NT.
 
 MIDI is not implemented yet.
*/


#include "m_imp.h"
#include <stdio.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <malloc.h>
#include <string.h>

int pa_open_audio(int inchans, int outchans, int rate, t_sample *soundin,
    t_sample *soundout, int advance);
void pa_close_audio(void);
int pa_send_dacs(void);
void pa_reportidle(void);

/* Defines */
#define DEBUG(x) x
#define DEBUG2(x) {x;}

#define PA_DEFAULTCH 2	    /* portaudio specific? */
typedef short t_pa_sample;
#define PA_SAMPLEWIDTH sizeof(t_pa_sample)
#define PA_BYTESPERCHAN (DACBLKSIZE * PA_SAMPLEWIDTH) 
#define PA_XFERSAMPS (PA_DEFAULTCH*DACBLKSIZE)
#define PA_XFERSIZE (PA_SAMPLEWIDTH * PA_XFERSAMPS)

static int mac_whichapi = API_PORTAUDIO;
static int mac_inchannels;
static int mac_outchannels;
static int mac_advance_samples; /* scheduler advance in samples */
static int mac_meters;    	/* true if we're metering */
static float mac_inmax;    	/* max input amplitude */
static float mac_outmax;    	/* max output amplitude */

    /* exported variables */
int sys_schedadvance = 50000; 	/* scheduler advance in microseconds */
float sys_dacsr;
int sys_hipriority = 0;
t_sample *sys_soundout;
t_sample *sys_soundin;


    /* set channels and sample rate.  */

static void mac_setchsr(int chin, int chout, int sr)
{
    int nblk;
    int inbytes = chin * (DACBLKSIZE*sizeof(float));
    int outbytes = chout * (DACBLKSIZE*sizeof(float));

    mac_inchannels = chin;
    mac_outchannels = chout;
    sys_dacsr = sr;
    mac_advance_samples = (sys_schedadvance * sys_dacsr) / (1000000.);
    if (mac_advance_samples < 3 * DACBLKSIZE)
    	mac_advance_samples = 3 * DACBLKSIZE;

    if (sys_soundin)
    	free(sys_soundin);
    sys_soundin = (t_float *)malloc(inbytes);
    memset(sys_soundin, 0, inbytes);

    if (sys_soundout)
    	free(sys_soundout);
    sys_soundout = (t_float *)malloc(outbytes);
    memset(sys_soundout, 0, outbytes);

    if (sys_verbose)
    	post("input channels = %d, output channels = %d",
    	    mac_inchannels, mac_outchannels);
}

/* ----------------------- public routines ----------------------- */

void sys_open_audio(int inchans, int outchans, int rate)
{
    int defaultchannels = PA_DEFAULTCH;

    if (rate < 1) rate = 44100;
    if (inchans < 0) inchans = defaultchannels;
    if (outchans < 0) outchans = defaultchannels;
    mac_setchsr(inchans, outchans, rate);
    pa_open_audio(inchans, outchans, rate, sys_soundin, sys_soundout,
    	mac_advance_samples);

}

void sys_close_audio(void)
{
    pa_close_audio();
}


int sys_send_dacs(void)
{
    if (mac_meters)
    {
    	int i, n;
	float maxsamp;
	for (i = 0, n = mac_inchannels * DACBLKSIZE, maxsamp = mac_inmax;
	    i < n; i++)
	{
	    float f = sys_soundin[i];
	    if (f > maxsamp) maxsamp = f;
	    else if (-f > maxsamp) maxsamp = -f;
	}
	mac_inmax = maxsamp;
	for (i = 0, n = mac_outchannels * DACBLKSIZE, maxsamp = mac_outmax;
	    i < n; i++)
	{
	    float f = sys_soundout[i];
	    if (f > maxsamp) maxsamp = f;
	    else if (-f > maxsamp) maxsamp = -f;
	}
	mac_outmax = maxsamp;
    }
    return pa_send_dacs();
}

float sys_getsr(void)
{
     return (sys_dacsr);
}

int sys_get_outchannels(void)
{
     return (mac_outchannels); 
}

int sys_get_inchannels(void) 
{
     return (mac_inchannels);
}

void sys_audiobuf(int n)
{
     /* set the size, in milliseconds, of the audio FIFO */
     if (n < 5) n = 5;
     else if (n > 5000) n = 5000;
     sys_schedadvance = n * 1000;
}

void sys_getmeters(float *inmax, float *outmax)
{
    if (inmax)
    {
    	mac_meters = 1;
	*inmax = mac_inmax;
	*outmax = mac_outmax;
    }
    else
    	mac_meters = 0;
    mac_inmax = mac_outmax = 0;
}

void sys_reportidle(void)
{
}

void sys_open_midi(int nmidiin, int *midiinvec,
    int nmidiout, int *midioutvec)
{
    if (nmidiin || nmidiout)
    	post("MIDI: not yet implemented");
}

void sys_close_midi( void)
{
}

void sys_putmidimess(int portno, int a, int b, int c)
{
}

void sys_putmidibyte(int portno, int byte)
{
}

void sys_poll_midi(void)
{
}

void sys_set_priority(int higher) 
{
#ifdef UNIX
    struct sched_param par;
    int p1 ,p2, p3;
#ifdef _POSIX_PRIORITY_SCHEDULING

    p1 = sched_get_priority_min(SCHED_FIFO);
    p2 = sched_get_priority_max(SCHED_FIFO);
    p3 = (higher ? p2 - 1 : p2 - 3);
    par.sched_priority = p3;

    if (sched_setscheduler(0,SCHED_FIFO,&par) != -1)
       fprintf(stderr, "priority %d scheduling enabled.\n", p3);
#endif

#ifdef _POSIX_MEMLOCK
    if (mlockall(MCL_FUTURE) != -1) 
    	fprintf(stderr, "memory locking enabled.\n");
#endif
#endif /* UNIX */
}

void nt_soundindev(int which)
{
}

void nt_soundoutdev(int which)
{
}

void nt_midiindev(int which)
{
}

void nt_midioutdev(int which)
{
}

void nt_noresync(void )
{
}

void nt_listdevs(void )
{
}

void glob_audio(void *dummy, t_floatarg fadc, t_floatarg fdac)
{
}
